#include <stdarg.h>
#include <stddef.h>
#include <inttypes.h>
#include <xen/xen.h>
#include <xen/callback.h>

#include "processor.h"
#include "shared.h"

struct regs_64 {
    /* pushed onto stack before calling into C code */
    uint64_t rax;
    uint64_t rbx;
    uint64_t rcx;
    uint64_t rdx;
    uint64_t r8;
    uint64_t r9;
    uint64_t r10;
    uint64_t r11;
    uint64_t r12;
    uint64_t r13;
    uint64_t r14;
    uint64_t r15;
    uint64_t rsi;
    uint64_t rdi;
    uint64_t rbp;
    uint64_t trapno;
    /* trap / fault / int created */
    uint64_t error;
    uint64_t rip;
    uint64_t cs;
    uint64_t rflags;
    uint64_t rsp;
    uint64_t ss;
};

/* 64bit defines */
#define longmode  1
#define regs      regs_64
#define emulate   emulate_64
#define fix_sel   fix_sel64
#define fix_desc  fix_desc64
#define ureg_t    uint64_t
#define sreg_t    int64_t
#define PRIxREG   PRIx64
#include "emu.h"

/* emu64.S */
extern void syscall(void);
extern char stack_low[];
extern char stack_high[];

/* emu-data.c */
extern struct descriptor_32 page_aligned xen_gdt[8192];
extern struct idt_64 page_aligned xen_idt[256];
extern struct tss_64 page_aligned xen_tss;
extern struct vcpu_guest_context page_aligned xen_ctxt;
extern struct shared_info page_aligned shared_info;

extern xen_callback_t       xencb[8];
extern struct trap_info     xentr[256];

extern uint64_t user_mode;
extern uint64_t kernel_cr3_mfn;
extern uint64_t user_cr3_mfn;

extern uint64_t gdt_mfn[16];
extern uint64_t emu_hcalls[XEN_HCALL_MAX];
extern uint64_t emu_mcalls[XEN_HCALL_MAX];

/* emu-main.c */
int panic(char *message);
int bounce_trap(struct regs_64 *regs, int trapno, int cbno);
asmlinkage void do_boot(struct regs_64 regs);
asmlinkage void do_illegal_instruction(struct regs_64 regs);
asmlinkage void do_general_protection(struct regs_64 regs);
asmlinkage void do_page_fault(struct regs_64 regs);
asmlinkage void do_event_callback(struct regs_64 regs);
asmlinkage void do_int_80(struct regs_64 regs);

/* emu-hcall.c */
void switch_mode(void);
int is_kernel(void);
asmlinkage void do_syscall(struct regs_64 regs);

/* emu-instr.c */
int emulate_64(struct regs_64 *regs);

/* emu-mm.c */
void pgtable_walk(uint64_t va, uint64_t root_mfn);
int pgtable_fixup_flag(uint64_t va, uint32_t flag);
void *map_64(uint64_t maddr);
uint64_t *find_pte_64(uint64_t va);
void update_emu_mappings_64(uint64_t cr3_mfn);

/* printk.c */
int vscnprintf(char *buf, size_t size, const char *fmt, va_list args);
int snprintf(char * buf, size_t size, const char *fmt, ...);
int printk(const char *fmt, ...) __attribute__((format(printf, 1, 2)));

/* macros */
#define context_is_emu(regs)    (((regs)->cs & 0x03) == 0x00)
#define context_is_kernel(regs) (((regs)->cs & 0x03) == 0x03 && is_kernel())
#define context_is_user(regs)   (((regs)->cs & 0x03) == 0x03 && !is_kernel())

