#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <xenctrl.h>
#include <dlfcn.h>
#include <sys/mman.h>

/* ------------------------------------------------------------------ */

static int           trace = 1;
static int           fixme = 1;

static int           evtpipe[2];
static evtchn_port_t ports = 105;

static evtchn_port_t xsd_port;
static int           xsd_fd;
static void          *xsd_map;

struct evtchn {
    struct evtchn *next;
    evtchn_port_t local_port;
    evtchn_port_t remote_port;
    uint32_t      remote_domain;
};
struct evtchn *evtchns;

/* ------------------------------------------------------------------ */

static struct evtchn *new_evtchn(void)
{
    struct evtchn *evtchn;

    evtchn = malloc(sizeof(*evtchn));
    memset(evtchn,0,sizeof(*evtchn));
    evtchn->next = evtchns;
    evtchns = evtchn;

    evtchn->local_port = ports++;
    return evtchn;
}

static struct evtchn *get_evtchn(evtchn_port_t local_port)
{
    struct evtchn *evtchn;

    for (evtchn = evtchns; NULL != evtchn; evtchn = evtchn->next)
	if (evtchn->local_port == local_port)
	    break;
    return evtchn;
}

static void del_evtchn(evtchn_port_t local_port)
{
    struct evtchn *this, *prev;

    for (prev = NULL, this = evtchns; NULL != this; prev = this, this = this->next)
	if (this->local_port == local_port)
	    break;
    if (!this) {
	fprintf(stderr, "libxc: %s: port %d not found\n",
		__FUNCTION__, local_port);
	return;
    }
    
    if (prev)
	prev->next = this->next;
    else
	evtchns = this->next;
    free(this);
}

/* ------------------------------------------------------------------ */

int xc_interface_open(void)
{
    int rc = 42;

    if (trace)
	fprintf(stderr, "libxc: %s -> %d\n", __FUNCTION__, rc);
    return rc;
}

int xc_interface_close(int xc_handle)
{
    if (trace)
	fprintf(stderr, "libxc: %s\n", __FUNCTION__);
    return 0;
}

int xc_domain_getinfo(int xc_handle,
                      uint32_t first_domid,
                      unsigned int max_doms,
                      xc_dominfo_t *info)
{
    if (fixme)
	fprintf(stderr, "libxc: %s: FIXME\n", __FUNCTION__);
    errno = ENOSYS;
    return -1;
}

/* ------------------------------------------------------------------ */

int xc_evtchn_open(void)
{
    static int first = 1;
    int rc = 69;

    if (!first) {
	fprintf(stderr, "libxc: %s called twice\n", __FUNCTION__);
	errno = ENOSYS;
	return -1;
    }
    first = 0;

    pipe(evtpipe);
    if (trace)
	fprintf(stderr, "libxc: %s -> handle %d\n",
		__FUNCTION__, rc);
    return rc;
}

int xc_evtchn_fd(int xce_handle)
{
    return evtpipe[0];
}

evtchn_port_t xc_evtchn_bind_interdomain(int xce_handle, int domid,
					 evtchn_port_t remote_port)
{
    struct evtchn *evtchn;

    if (domid == 0) {
	if (remote_port != xsd_port)
	    goto fail;
    } else {
	/* TODO: check domid exists */;
    }

    evtchn = new_evtchn();
    evtchn->remote_domain = domid;
    evtchn->remote_port = remote_port;
    if (trace)
	fprintf(stderr, "libxc: %s(handle %d, domid %d, rport %d) -> port %d\n",
		__FUNCTION__, xce_handle, domid, remote_port, evtchn->local_port);
    return evtchn->local_port;

fail:
    errno = EINVAL;
    return -1;
}

evtchn_port_t xc_evtchn_bind_virq(int xce_handle, unsigned int virq)
{
    struct evtchn *evtchn;

    evtchn = new_evtchn();
    if (fixme)
	fprintf(stderr, "libxc: %s(handle %d, virq %d) -> port %d, FIXME\n",
		__FUNCTION__, xce_handle, virq, evtchn->local_port);
    return evtchn->local_port;
}

int xc_evtchn_unbind(int xce_handle, evtchn_port_t port)
{
    if (trace)
	fprintf(stderr, "libxc: %s(handle %d, port %d)\n",
		__FUNCTION__, xce_handle, port);
    del_evtchn(port);
    return 0;
}

int xc_evtchn_notify(int xce_handle, evtchn_port_t port)
{
    struct evtchn *evtchn;

    evtchn = get_evtchn(port);
    if (NULL == evtchn)
	goto fail;

    if (evtchn->remote_port) {
	if (fixme)
	    fprintf(stderr, "libxc: %s(handle %d, port %d)"
		    " -> domain %d, port %d, FIXME\n",
		    __FUNCTION__, xce_handle, port,
		    evtchn->remote_domain, evtchn->remote_port);
	/* FIXME */
    } else {
	if (fixme)
	    fprintf(stderr, "libxc: %s(handle %d, port %d) -> FIXME\n",
		    __FUNCTION__, xce_handle, port);
	/* FIXME */
    }
    return 0;

fail:
    fprintf(stderr, "libxc: %s(handle %d, port %d) -> invalid port\n",
	    __FUNCTION__, xce_handle, port);
    errno = EINVAL;
    return -1;
}

evtchn_port_t xc_evtchn_pending(int xce_handle)
{
    if (fixme)
	fprintf(stderr, "libxc: %s: FIXME\n", __FUNCTION__);
    errno = ENOSYS;
    return -1;
}

int xc_evtchn_unmask(int xce_handle, evtchn_port_t port)
{
    if (fixme)
	fprintf(stderr, "libxc: %s: FIXME\n", __FUNCTION__);
    errno = ENOSYS;
    return -1;
}

/* ------------------------------------------------------------------ */

void *xc_map_foreign_range(int xc_handle, uint32_t dom,
			   int size, int prot,
			   unsigned long mfn)
{
    if (fixme)
	fprintf(stderr, "libxc: %s(handle %d, dom %d, size %d, prot %d, mfn 0x%lx):"
		" FIXME: fake mapping\n",
		__FUNCTION__, xc_handle, dom, size, prot, mfn);
    
    /* FIXME: fake mapping */
    return mmap(NULL, size, prot, MAP_SHARED | MAP_ANON, -1, 0);
}

/* ------------------------------------------------------------------ */

static int fake_xsd_port(void)
{
    char file[32] = "/tmp/xsd_port.XXXXXXXX";
    char buf[16];
    int fd;

    xsd_port = ports++;
    if (trace)
	fprintf(stderr, "libxc: %s, port %d\n", __FUNCTION__, xsd_port);
    snprintf(buf, sizeof(buf), "%d\n", xsd_port);
    fd = mkstemp(file);
    write(fd, buf, strlen(buf));
    lseek(fd, SEEK_SET, 0);
    unlink(file);
    return fd;
}

static int fake_xsd_kva(void)
{
    char file[32] = "/tmp/xsd_kva.XXXXXXXX";

    if (trace)
	fprintf(stderr, "libxc: %s\n", __FUNCTION__);
    xsd_fd = mkstemp(file);
    lseek(xsd_fd, SEEK_SET, PAGE_SIZE-1);
    write(xsd_fd, '\0', 1);
    lseek(xsd_fd, SEEK_SET, 0);
    xsd_map = mmap(NULL, PAGE_SIZE, PROT_READ | PROT_WRITE,
		   MAP_SHARED, xsd_fd, 0);
    unlink(file);
    return xsd_fd;
}

/* open syscall wrapper, to fake /proc/xen files */
int open(const char *file, int flags, mode_t mode);
int open(const char *file, int flags, mode_t mode)
{
    static int (*real_open)(const char *file, int flags, mode_t mode);

    if (!real_open)
	real_open = dlsym(RTLD_NEXT, "open");

    if (0 == strcmp(file, "/proc/xen/xsd_port"))
	return fake_xsd_port();
    if (0 == strcmp(file, "/proc/xen/xsd_kva"))
	return fake_xsd_kva();

    /* default: forward */
    if (trace)
	fprintf(stderr, "libxc: %s: %s\n", __FUNCTION__, file);
    return real_open(file, flags, mode);
}
