#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <inttypes.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/mman.h>

#include "xenner.h"

/* ------------------------------------------------------------------ */

#define VMCORE_HEADER (2 * 1024 * 1024)   /* 2MB page */

static const char *corepath[] = {
    "/var/run/xenner",
    "."
};

int vmcore_create(struct xenvm *xen)
{
    int i;

    for (i = 0; i < sizeof(corepath)/sizeof(corepath[0]); i++) {
	snprintf(xen->corefile, sizeof(xen->corefile), "%s/vmcore.%d",
		 corepath[i], xen->domid);
	xen->corefd = open(xen->corefile, O_CREAT | O_EXCL, 600);
	if (xen->corefd >= 0)
	    break;
	if (EEXIST == errno)
	    goto fail;
    }
    if (i == sizeof(corepath)/sizeof(corepath[0]))
	goto fail;

    /* create and map sparse file */
    xen->coresize = VMCORE_HEADER + 4096 * xen->pg_total;
    lseek(xen->corefd, xen->coresize -1, SEEK_SET);
    write(xen->corefd, "\0", 1);
    xen->coremap = mmap(NULL, xen->coresize, PROT_READ | PROT_WRITE,
			MAP_SHARED, xen->corefd, 0);
    xen->memory = xen->coremap + VMCORE_HEADER;
    return 0;

fail:
    fprintf(stderr, "open %s: %s\n", xen->corefile, strerror(errno));
    return -1;
}
